#ifndef lint
static char *rcsid = "$Header: /iss/planning/idi/lim-clos-version/CVSROOT/c-kqml/facilitator.c,v 1.2 1995/06/23 20:49:41 robin Exp $";
#endif


/*
 * $Log: facilitator.c,v $
 * Revision 1.2  1995/06/23  20:49:41  robin
 * Added a persistent store to the ANS so that if the ANS goes
 * down, it will restore it's agent cache with what it had before it went
 * down.
 *
 * Revision 1.3  1993/04/30  19:28:58  fritzson
 * Corrected infinite loop with debug turned off.
 *
 * Revision 1.2  1993/01/26  22:04:32  robin
 * Modified to put application debug messages to separate file and to allow
 * user to pass in router debug level from the command line.
 *
 * Revision 1.1  1993/01/25  13:47:02  fritzson
 * Initial revision
 *
 *
 */


/*
 *	(c) Copyright 1993 Loral Defense Systems-Eagan
 *
 *	This material my be reproduced by or for the U.S.
 *	Government pursuant to the copyright license under
 *	the clause at DFARS 252.227-7013   (1991).
 */

#define	MAX_HOSTNAME_LEN	256

/* KQML ANS 
 *
 * A first pass at a simple ANS.
 *
 * Accepts the following types of KQML performatives:
 *
 * (register   :content (symbolic_name host port)) and replies with
 * (reply      :content symbolic_name)

 * (unregister :content (symbolic_name host port)) 

 * (ask-one    :content "symbolic_name") and replies with
 * (tell :content "symbolic_name:host:port:" ) 

 * (ask-all    :content "symbolic_name" ) and replies with
 * (tell :content ((symbolic_name (protocol host:port:)) 
 *      (symbolic-name (protocol host:port:))))
 * if name is a proper symbolic_name, it is equivalent to ask-one
 * if name is not defined or is *, all the entries are returned 
 *
 * It is implemented in the crudest possible way, maintaining 
 * an unordered array of entries where each entry contains a service's
 * symbolic name, its network information
 * 
 * Note that this version only supports one instance of a symbolic
 * name in the database at a time.
 */

#include <kqml/kqmlconf.h>

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/time.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ndbm.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <memory.h>

#include <kqml.h>
#include <kqml/kqmllib.h>

#ifdef __KQML_MACH_SUN__
	#include <malloc.h>
#endif

#ifdef __KQML_MACH_SVR4__
	#include <sys/systeminfo.h>
#endif

struct net_list
{
	char *protocol;
	char *addr_info;
	struct net_list *prev;
	struct net_list *next;
	struct net_list *match;
};

/* The linear table of entries */
struct knet
{
	char *name;
	struct net_list *net_list;
	struct net_list *net_match;
	struct knet *prev;
	struct knet *next;
	struct knet *match;	/* Keep track of selected entries */
};

extern char *optarg;
extern int optind;

/* Forward declarations */
static int ans_name_lookup (char *agent_name, char **name, char **protocol, 
	char **addr_info);
static int agent_is_alive (char *agent_name);
static void deliver_info (int flag, struct knet *node);

static void delete_dbm_knet_entry (struct knet *node);
static void remove_net_list_entry (int dbm_delete, struct knet *node);
static void remove_knet_entry (int dbm_delete, struct knet *node);
static void free_net_list_entry (struct net_list *entry);
static void free_knet_entry (struct knet *entry);

static struct knet *add_knet_entry (struct knet *node, char *name, 
	char *protocol, char *addr_info);
static int init_knet_table ();

static int read_info (char *protocol, char *message, char **addr_info);
static int compare_info (char *protocol, char *addr1, char *addr2);

static int form_dbm_content (struct knet *node, datum *content);

static void *kqml_admin_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type);
static void *ask_one_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type);
static void *ask_all_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type);
static void *register_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type);
static void *unregister_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type);

static struct knet *find_exact_agent (char *name, char *protocol);

static struct net_list *find_any_proto (struct knet *node, char *proto, 
	int multi_flag, int *match_len);
static struct knet *find_any_agent (char *name, char *protocol, 
	int multi_flag, int *match_len);
static char *form_agents_list (struct knet *entries, int multi_flag, 
	int length);

static struct knet *knet_table = NULL;
static int knet_tbl_cnt = 0;

/* Global variables */
static char *dbm_file="fac_db";  
static char *dbm_pag_file="fac_db.pag";  
static int remote_shutdown_enabled = FALSE;

/*
 *  For an agent on a given host at a given port, check to see if that agent 
 *	 is still alive.
 *
 */

static int
ans_name_lookup (char *agent_name, char **name, char **protocol, 
	char **addr_info)
{
	struct knet *node;

	/* 
	 * Currently, i am using tcp-ip as the protocol, later 
	 * the ans can do something more intelligent based on
	 * the protocols supported by its router. 
	 */

	node = find_exact_agent (agent_name, "tcp-ip");
	if (node == NULL || node->net_match == NULL)
	{
		/* 
		 * If we do not find an agent we should add code to return
		 * the address of somebody who might know how to find that 
		 * agent. Heirarchical ANS ;-)
		*/

		return FALSE;
	}

	*name = node->name;
	*protocol = node->net_match->protocol;
	*addr_info = node->net_match->addr_info;

	return TRUE;
}

static int
agent_is_alive (char *agent_name)
{
	char *reply_perf, *reply_name;
	kqml_message *rmsg;
	int ret_value;

	/* for now, bypass this processing */
	/* return FALSE; */

	/* Ask this agent who it is (to make sure it's who I think it should be) */
	ret_value = kqml_send_msg (KQML_SEND_TIMEOUT, "kqml-admin", "(name)",
		agent_name, "t", &rmsg, NULL);

	if (ret_value <= -1)
	{
		fprintf (stderr, "agent %s is no longer there\n", agent_name);
		fflush (stderr);
		return FALSE;
	}

	/* Was i able to reach an agent at this address? */
	reply_perf = kqml_get_performative (rmsg);
	if (kqml_stricmp ("error", reply_perf) == 0)
	{
		fprintf (stderr, "Unable to reach agent [%s] at this address\n", 
			agent_name);
		fflush (stderr);
		return FALSE;
	}

	if (kqml_stricmp ("sorry", reply_perf) == 0)
	{
		fprintf (stderr, 
			"Agent [%s] unable to provide response to name inquiry\n", agent_name);
		fflush (stderr);
		return FALSE;
	}

	reply_name = kqml_get_field (":content", rmsg, NULL, NULL);
	if (kqml_stricmp ("reply", reply_perf) != 0 || reply_name == NULL)
	{
		fprintf (stderr, 
			"Agent [%s] provided illegal response to name inquiry\n", agent_name);
		fflush (stderr);
		return FALSE;
	}

	/* There is an agent at that address, but check to see if it's the one 
		I'm expecting */
	if (kqml_stricmp (reply_name, agent_name) == 0)
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}

static void
deliver_info (int flag, struct knet *node)
{
	kqml_message *msg;
	char *protocol, *addr_info, *deliver_str;

	if (node->net_match)
	{
		protocol = node->net_match->protocol;
		addr_info = node->net_match->addr_info;
	}
	else
	{
		protocol = addr_info = "";
	}

	msg = kqml_build_msg ("transport-address", addr_info,
		NULL, NULL, ":name", node->name, ":ontology", protocol, NULL);

	deliver_str = kqml_sprint_msg (NULL, NULL, msg);
	kqml_free_msg (msg);

	if (deliver_str)
	{
		kqml_deliver_msg ((flag ? "tell" : "untell"), deliver_str, NULL);
		free (deliver_str);
	}
}

static void
delete_dbm_knet_entry (struct knet *node)
{
	datum dbm_key;
	DBM *fac_db;

	if (node == NULL)
	{
		return ;
	}

	dbm_key.dptr  = node->name;
	dbm_key.dsize = strlen (dbm_key.dptr) + 1;

	fac_db = dbm_open (dbm_file, O_WRONLY, 0);
	if (fac_db == NULL)
	{
		fprintf (stderr, 
			"Unable to open DBM database %s in delete entry\n", dbm_file);
		exit (1);
	}

	/* delete the entry from the dbm database */
	if ( dbm_delete (fac_db, dbm_key) < 0 )
	{
		fprintf (stderr,
			"Unable to delete DBM entry '%s' in delete entry\n", dbm_key.dptr);
	}

	/* close the database to commit the modification */
	dbm_close (fac_db);
}

static void
remove_net_list_entry (int dbm_delete, struct knet *node)
{
	datum key, content;
	DBM *fac_db;

	if (node->net_match == NULL)
	{
		return;
	}

	/* Critical section start ????? */
	if (node->net_list == node->net_match)
	{
		if (node->net_list == node->net_list->next)
		{
			node->net_list = NULL;
		}
		else
		{
			node->net_list = node->net_list->next;
		}
	}

	node->net_match->prev->next = node->net_match->next;
	node->net_match->next->prev = node->net_match->prev;
	/* Critical section end   ????? */

	if (dbm_delete)
	{
		key.dptr = node->name;
		key.dsize = strlen (node->name) + 1;
		if (form_dbm_content (node, &content))
		{
			fac_db = dbm_open (dbm_file, O_RDWR, 0);
			if (fac_db == NULL)
			{
				fprintf (stderr, 
					"Unable to open DBM database %s in remove_net_list_entry \n",
					dbm_file);
				fflush (stderr);
				exit (1);
			}

			if (dbm_store (fac_db, key, content, DBM_REPLACE) <= -1)
			{
				fprintf (stderr, 
					"Unable to store new value, '%s', in remove_net_list_entry\n", 
					content.dptr);
			}

			dbm_close (fac_db);

			free (content.dptr);
		}
	}

	free_net_list_entry (node->net_match);

	node->net_match = NULL;

	return;
}

static void
remove_knet_entry (int dbm_delete, struct knet *node)
{
	if (node == NULL)
	{
		return ;
	}

	if (dbm_delete)
	{
		/* delete the entry from the persistent store */
		delete_dbm_knet_entry (node);
	}
  
	/* Critical section  start ????? */
	if (node == node->next)
	{
		knet_table = NULL;
	}
	else if (node == knet_table)
	{
		knet_table = node->next;
	}

	node->prev->next = node->next;
	node->next->prev = node->prev;
	knet_tbl_cnt--;
	/* Critical section  end   ????? */

	free_knet_entry (node);
}

static void
free_net_list_entry (struct net_list *entry)
{
	if (entry)
	{
		if (entry->protocol)
		{
			free (entry->protocol);
		}
		if (entry->addr_info)
		{
			free (entry->addr_info);
		}
		free (entry);
	}
}

static void 
free_knet_entry (struct knet *entry)
{
	struct net_list *temp, *next;

	if (entry)
	{
		if (entry->name)
		{
			free (entry->name);
		}
		if (entry->net_list)
		{
			temp = entry->net_list;
			do
			{
				next = temp->next;
				free_net_list_entry (temp);
				temp = next;
			}
			while (temp != entry->net_list); 
		}
		free (entry);
	}
}

static struct knet *
add_knet_entry (struct knet *node, char *name, char *protocol, char *addr_info)
{
	struct knet *new_entry;
	struct net_list *temp;
	char *addr_info_copy;

	addr_info_copy = strdup (addr_info);
	if (addr_info_copy == NULL)
	{
		return NULL;
	}

	if (node != NULL && node->net_match != NULL)
	{
		new_entry = node;
		if (new_entry->net_match->addr_info)
		{
			free (new_entry->net_match->addr_info);
		}
		new_entry->net_match->addr_info = addr_info_copy;
	}
	else
	{
		temp = malloc (sizeof (struct net_list));
		if (temp == NULL)
		{
			free (addr_info_copy);
			return NULL;
		}
				
		temp->prev = temp->next = temp;
		temp ->addr_info = addr_info_copy;
		temp ->protocol = strdup (protocol);
		if (temp ->protocol == NULL)
		{
			free_net_list_entry (temp);
			return NULL;
		}

		if (node == NULL)
		{
			new_entry = malloc (sizeof (struct knet));
			if (new_entry == NULL)
			{
				free_net_list_entry (temp);
				return NULL;
			}

			new_entry->prev = new_entry->next = new_entry->match = NULL;
			new_entry->net_list = new_entry->net_match = NULL;

			new_entry->name = strdup (name);
			if (new_entry->name == NULL)
			{
				free_knet_entry (new_entry);
				free_net_list_entry (temp);
				return NULL;
			}
		}
		else
		{
			new_entry = node;
		}

		new_entry->net_match = temp;

		/* critical section start ???? */
		if (new_entry->net_list)
		{
			temp->next = new_entry->net_list;
			temp->prev = new_entry->net_list->prev;
			new_entry->net_list->prev->next = temp;
			new_entry->net_list->prev = temp;
		}
		else
		{
			new_entry->net_list =
				temp->next =
				temp->prev = temp;
		}
		/* critical section end   ???? */
	}

	if (node == NULL)
	{
		/* Critical section start ????? */
		if (knet_table == NULL)
		{
			knet_table = 
				new_entry->next =
				new_entry->prev = new_entry;
			knet_tbl_cnt = 1;
		}
		else
		{
			new_entry->next = knet_table;
			new_entry->prev = knet_table->prev;
			knet_table->prev->next = new_entry;
			knet_table->prev = new_entry;
			knet_tbl_cnt++;
		}
		/* Critical section end   ????? */
	}

	return new_entry;
}

/*
 * init_knet_table
 *
 * Initialize the table of all entries in the ANS's data store.  
 *	This includes registering the ANS itself and any entries that 
 *	have been stored in it's dbm data store (for persistence).
 *
 */

static int
init_knet_table ()
{
	datum key, ans;
	DBM *fac_db;
	struct knet *node;
	char *temp, *protocol, *addr_info;
	int len;
	struct stat stat_buf;
  
	knet_tbl_cnt = 0;
	knet_table = NULL;

	if (add_knet_entry (NULL, kqml_agent_name, kqml_protocol, 
		kqml_addr_info) == NULL)
	{
		fprintf (stderr, "Unable to create an entry for self\n");
		return -1;
	}

  /* Check for the existence of the DBM database */
	if ( (stat (dbm_pag_file, &stat_buf)) < 0)  
	{
#		ifdef DEBUG
			fprintf (stderr, "ANS database does not yet exist\n");
			fflush (stderr);
#		endif

		/* No DBM file exists yet... I'll create one */
		fac_db = dbm_open (dbm_file, O_CREAT, 0664);
		if (fac_db == NULL)
		{
			fprintf (stderr, "Unable to open ANS database %s for create\n",
				dbm_file);
			fflush (stderr);
			exit (1);
		}
		dbm_close (fac_db);
	}
	else
	{
#		ifdef DEBUG
			fprintf (stderr, "ANS database exists,");
			fprintf (stderr, "and will be used to refresh connections table\n");
			fflush (stderr);
#		endif
	}
  
	fac_db = dbm_open (dbm_file, O_RDWR, 0);
	if (fac_db == NULL)
	{
		fprintf (stderr, "Unable to open DBM database %s\n", dbm_file);
		fflush (stderr);
		exit (1);
	}

	/* read in the database entries and store them in the knet_table */
	for (key=dbm_firstkey(fac_db); key.dptr != NULL; key = dbm_nextkey(fac_db))
	{ 
		ans = dbm_fetch (fac_db, key);

		temp = ans.dptr;
		len = ans.dsize;
		node = NULL;
		while (*temp)
		{
			protocol = temp;
			temp = memchr (temp, 0, len);
			addr_info = ++temp;
			temp = memchr (temp, 0, len);
			++temp;

      	/* Copy valid new entry into the connections table */
#			ifdef DEBUG
				fprintf (stderr, "KNET Entry: %s %s %s\n", key.dptr, 
					protocol, addr_info);
				fflush (stderr);
#			endif

			node = add_knet_entry (node, key.dptr, protocol, addr_info);
			if (node == NULL)
			{
				fprintf (stderr, "Could not add old Knet entry %s \n",
					key.dptr);
				break;
			}
			
			deliver_info (TRUE, node);
			node->net_match = NULL;
		}

		if (! agent_is_alive (key.dptr))
		{
			node->net_match = NULL;
			deliver_info (FALSE, node);
			remove_knet_entry (TRUE, node);
		}
	}

	dbm_close (fac_db);

	return TRUE;
}


static int
read_info (char *protocol, char *message, char **addr_info)
{
	int len;
	char *temp;

	if (message == NULL)
	{	
		return -1;
	}

	*addr_info = NULL;

	if (kqml_stricmp (protocol, "tcp-ip") == 0)
	{
		char *host, *port_str;
		int port;

		host = strtok (message, ": \t\n");	/* get host name */
		if (host == NULL)
		{
			fprintf (stderr, "  Couldn't parse host.\n");
			return -1;
		}

		port_str = strtok (NULL, ": \t\n)");	/* get port number */
		if (port_str == NULL)
		{
			fprintf (stderr, "  Couldn't parse port.\n");
			return -1;
		}
		port = atoi (port_str);
		if (port == 0)
		{
			fprintf (stderr, "  Port wasn't a number.\n");
			return -1;
		}

		/* form the addr_info string */
		len = strlen (host) + strlen (port_str) + 3;
		*addr_info = malloc (len);

		if (*addr_info == NULL)
		{
			fprintf (stderr, "  Could not form addr_info\n");
			return -1;
		}

		temp = memccpy (*addr_info, host, 0, len);
		*(temp - 1) = ':';
		/* Do not use port_str. We might have trouble in comparing two
			addr_info's */
		temp += kqml_digit_to_str (port, temp);
		*temp++ = ':';
		*temp = '\0';

		return 0;
	}
	else if (kqml_stricmp (protocol, "smtp") == 0 ||
		(kqml_stricmp (protocol, "http") == 0))
	{
		char *temp;

		/* Remove the leading spaces */
		while (isspace (*message))
			message++;

		temp = strtok (message, " \t\n)");	/* get addr_info */
		if (temp == NULL)
		{
			fprintf (stderr, "  Couldn't read address.\n");
			return -1;
		}

		*addr_info = strdup (temp);
		if (*addr_info == NULL)
		{
			fprintf (stderr, "  Could not form addr_info\n");
			return -1;
		}

		return 0;
	}
	else
	{
		return -1;
	}
}

static int
compare_info (char *protocol, char *addr1, char *addr2)
{
	if (kqml_stricmp (protocol, "tcp-ip") == 0 ||
		kqml_stricmp (protocol, "smtp") == 0 ||
		kqml_stricmp (protocol, "http") == 0)
	{
		return kqml_stricmp (addr1, addr2) == 0;
	}
	else
		return FALSE;
}

static int
form_dbm_content (struct knet *node, datum *content)
{
	char *temp_content;
	struct net_list *temp;
	int len;
	
	if (node->net_list == NULL)
	{
		return FALSE;
	}

	len = 0;
	temp = node->net_list;
	do
	{
		len += strlen (temp->protocol) + strlen(temp->addr_info) + 2;
		temp = temp->next;
	}
	while (temp != node->net_list);

	content->dsize = len + 1;
	content->dptr = malloc (content->dsize);

	if (content->dptr == NULL)
	{
		return FALSE;
	}

	temp_content = content->dptr;
	temp = node->net_list;
	do
	{
		temp_content = memccpy (temp_content, temp->protocol, 0, len);
		temp_content = memccpy (temp_content, temp->addr_info, 0, len);
		temp = temp->next;
	}
	while (temp != node->net_list);
	
	*temp_content = '\0';

	return TRUE;
}


static void *
kqml_admin_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type)
{
	char *content;
	kqml_message *msg;

	*reply_type = KQML_REPLY_ERROR;

	content = kqml_get_field (":content", kmsg, NULL, NULL);
	if (content == NULL)
	{
		fprintf (stderr, "Received empty kqml-admin message.\n");
		return NULL;
	}

	content = strchr (content, '(');
	if (content == NULL)
	{
		fprintf (stderr, "Received empty content in kqml-admin.\n");
		return NULL;
	}
	content++;

	content = strtok (content, " \t\n)");
	if (content == NULL)
	{
		fprintf (stderr, "No command in kqml-admin.\n");
		return NULL;
	}

	if (remote_shutdown_enabled)
	{
		if (kqml_stricmp (content, "shutdown") == 0)
		{
			exit (0);
		}
	}

	fprintf (stderr, "Unsupported command in kqml-admin.\n");
	msg = kqml_build_msg ("sorry", "Not supported", NULL, NULL, NULL);
	*reply_type = KQML_REPLY_KMSG;
	return msg;
}

/*
 * ask_one_handler
 *
 * Return a string containing the full entry for the given host, or 
 * an empty string if there is no entry.
 *
 */
static void *
ask_one_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type)
{
	kqml_message *msg;
	struct knet *node;
	char *protocol;
	int len;

#	ifdef DEBUG
		fprintf (stderr, "ask_one_handler: '%s'\n", message);
#	endif

	if (message == NULL)
	{
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}

	protocol = kqml_get_field (":ontology", kmsg, NULL, NULL);
	if (protocol == NULL)
	{
		protocol = "tcp-ip";
	}

	node = find_any_agent (message, protocol, FALSE, &len);
	if (node == NULL || node->net_match == NULL)
	{
#		ifdef DEBUG
			fprintf (stderr, "  NOT FOUND\n");
#		endif

		msg = kqml_build_msg ("sorry", NULL, NULL, NULL, NULL);
	}
	else 
	{
		char *reply;

#		ifdef DEBUG
			fprintf (stderr, "  FOUND as %s %s %s\n",	node->name, 
				node->net_match->protocol,
				node->net_match->addr_info);
#		endif

		reply = form_agents_list (node, FALSE, len);
		if (reply == NULL)
		{
			fprintf (stderr, " Could not construct content.\n");
			*reply_type = KQML_REPLY_ERROR;
			return NULL;
		}

		msg = kqml_build_msg ("reply", reply, NULL, NULL, NULL);
		free (reply);
		if (msg == NULL)
		{
			fprintf (stderr, " Could not construct reply message.\n");
			*reply_type = KQML_REPLY_ERROR;
			return NULL;
		}

	}

	*reply_type = KQML_REPLY_KMSG;
	return msg;
}


/*
 * ask_all_handler
 *
 * Return a Lisp list of strings containing the full entry for the given
 * host, or an empty string if there are no entries.
 *
 */

static void *
ask_all_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type)
{
	kqml_message *msg;
	struct knet *node;
	int len;
	char *protocol;

#	ifdef DEBUG
		fprintf (stderr, "ask_all_handler: '%s'\n", message);
		fflush (stderr);
#	endif

	protocol = kqml_get_field (":ontology", kmsg, NULL, NULL);

	node = find_any_agent (message, protocol, TRUE, &len);
	if ( node == NULL )
	{
#		ifdef DEBUG
			fprintf (stderr, "  NOT FOUND\n");
#		endif

		msg = kqml_build_msg ("sorry", NULL, NULL, NULL, NULL);
	}
	else 
	{
		char	*reply;

		reply = form_agents_list (node, TRUE, len);
#		ifdef DEBUG
			fprintf (stderr, "  return from form_agents_list with '%s'\n", 
				reply);
#		endif

		if (reply == NULL)
		{
			fprintf (stderr, " Could not construct content.\n");
			*reply_type = KQML_REPLY_ERROR;
			return NULL;
		}

		msg = kqml_build_msg ("reply", reply, NULL, NULL, NULL);
		free (reply);
	}

	*reply_type = KQML_REPLY_KMSG;
	return msg;
}


/*
 * register_handler
 *
 * Add a new entry to the array.
 *
 */
static void *
register_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type)
{
	char *name, *protocol, *addr_info, *reply;
	struct knet *node;
	int reply_len, len, registered;
	datum key, content;
	DBM *fac_db;
	kqml_message *msg;
  
#	ifdef DEBUG
		fprintf (stderr, "Handling Register for %s\n", message);
		fflush (stderr);
#	endif

	protocol = kqml_get_field (":ontology", kmsg, NULL, NULL);
	if (protocol == NULL)
	{
		protocol = "tcp-ip";
	}

	/* cannot use strtok to remove the leading white spaces */
	message = strchr (message, '(');
	if (message == NULL)
	{
		fprintf (stderr, "  Couldn't find leading paren.\n");
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}
	message++;

	/* Extract symbolic name from the message */
	name = strtok (message, " \t\n)");
	
	if (name == NULL)
	{
		fprintf (stderr, "  Couldn't parse name.\n");
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}

	len = strlen (name);
	if (_kqml_check_word (name, len) != 0 || memchr (name, '.', len) != NULL)
	{
		fprintf (stderr, "  Illegal characters in agent name.\n");
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}

	if (read_info (protocol, strtok (NULL, ")"), &addr_info) <= -1)
	{
		fprintf (stderr, "  Could not parse addr_info.\n");
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}

	/* Check for this being a duplicate */
#	ifdef DEBUG
		fprintf (stderr, "  Checking for a duplicate entry.\n");
#	endif

	registered = FALSE;
	node = find_exact_agent (name, protocol);
	if (node != NULL)
	{
		if (node->net_match != NULL)
		{
			/* There is an agent with that name using the same protocol */
			if (compare_info (protocol, addr_info, node->net_match->addr_info))
			{
				/* If it is the same address, we can ignore it.
					It is a duplicate registration by the same agent. */
				free (addr_info);
				registered = TRUE;
			}
			else
			{
				/* Somebody is trying to use this name. Check to make sure
					that the other agent is alive and if so, reject this 
					request.
				*/
				if (agent_is_alive (node->name))
				{
					fprintf (stderr, "  Not adding duplicate entry; %s %s\n", 
						name, addr_info);
					free (addr_info);
					*reply_type = KQML_REPLY_ERROR;
					return NULL;
				}
			}
		}
	}
	else
	{
		/* No entry in that name using that protocol */
	}

	if (! registered)
	{
		/* Add fully qualified agent name */
		node = add_knet_entry (node, name, protocol, addr_info);

		if (node == NULL)
		{
			fprintf (stderr, "  Could not add entry; %s %s\n",
				name, addr_info);
			free (addr_info);
			*reply_type = KQML_REPLY_ERROR;
			return NULL;
		}
		else
		{
			free (addr_info);
		}
	}

	if (! registered && kqml_stricmp (node->name, kqml_agent_name) != 0)
	{
		/* If it is not registered and if it is not our entry */
		/* store the entry in the database, for persistence */
		key.dptr  = node->name;
		key.dsize = strlen (key.dptr) + 1;

		if (form_dbm_content (node, &content))
		{
			fac_db = dbm_open (dbm_file, O_RDWR, 0);
			if (fac_db == NULL)
			{
				fprintf (stderr, 
					"Unable to open DBM database %s in register_handler\n",
					dbm_file);
				fflush (stderr);
				exit (1);
			}

			if (dbm_store (fac_db, key, content, DBM_REPLACE) <= -1)
			{
				fprintf (stderr, 
					"Unable to store new value, '%s', in register_handler\n", 
					content.dptr);
			}

			dbm_close (fac_db);

			free (content.dptr);
		}

		deliver_info (TRUE, node);
#		ifdef DEBUG
			fprintf (stderr, "  Added entry for name= %s, addr_info= %s\n",
				node->name, node->net_match->addr_info);
#		endif
 	}

	reply_len = strlen (node->name) + strlen (kqml_agent_name) + 4;
	reply = malloc (reply_len);
	if (reply)
	{
		char *temp;

		temp = reply;
		*temp++ = '(';
		temp = memccpy (temp, node->name, 0, reply_len);
		*(temp - 1) = ' ';
		temp = memccpy (temp, kqml_agent_name, 0, reply_len);
		*(temp - 1) = ')';
		*temp = '\0';

		msg = kqml_build_msg ("reply", reply, NULL, NULL, NULL);
		free (reply);
		*reply_type = KQML_REPLY_KMSG;
		return msg;
	}
	else
	{
		fprintf (stderr, " No space for reply message in register_handler.\n");
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}
}


/* 
 * unregister_handler
 *
 * Remove an entry from the set.
 *
 */
static void *
unregister_handler (char *message, kqml_message *kmsg, 
	KQML_REPLY_TYPE *reply_type)
{
	char *name, *protocol;
	struct knet *node;
  
#	ifdef DEBUG
		fprintf (stderr, "Handling Unregister for %s\n", message);
		fflush (stderr);
#	endif

	if (message == NULL)
	{
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}

	protocol = kqml_get_field (":ontology", kmsg, NULL, NULL);

	if (message[0] == '(')
	{
		message++;
	}

	/* Extract symbolic name from the message */
	name = strtok (message, " \t\n)");
	if (name == NULL)
	{
		fprintf (stderr, "  No name in unregister message.\n");
		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}

	/* search for the entry */
	node = find_exact_agent (name, protocol);
	if (node == NULL || (protocol != NULL && node->net_match == NULL))
	{
#		ifdef DEBUG
      	fprintf (stderr, "  No entry for %s\n", name);
      	fflush (stderr);
#		endif

		*reply_type = KQML_REPLY_ERROR;
		return NULL;
	}

	if (protocol == NULL || node->net_list->next == node->net_list)
	{
		/* If this is the only entry or we recd. instructions to clean
			up this agent */
		/* Inform other's to clean up this */
		node->net_match = NULL;
		deliver_info (FALSE, node);
		if (kqml_stricmp (kqml_agent_name, node->name) != 0)
		{
			remove_knet_entry (TRUE, node);
		}
		else
		{
			/* If it is our entry, it will not be in database */
			remove_knet_entry (FALSE, node);
		}
	}
	else
	{
		/* Inform router and other's to remove this protocol */
		deliver_info (FALSE, node);
		if (kqml_stricmp (kqml_agent_name, node->name) != 0)
		{
			remove_net_list_entry (TRUE, node);
		}
		else
		{
			/* If it is our entry, it will not be in database */
			remove_net_list_entry (FALSE, node);
		}
	}

#	ifdef DEBUG
		fprintf (stderr, "  Removed entry for %s, new count is %d\n", message, 
			knet_tbl_cnt);
#	endif

	*reply_type = KQML_REPLY_NONE;
	return NULL;
}


int main (int argc, char **argv)
{
	char *temp;
	int	c;			/* for handling command line args */

	/* process command line arguments */
	while ((c = getopt (argc, argv, "d:r")) != -1)
		switch (c)
		{
			case 'd':
				break;

			case 'r':
				remote_shutdown_enabled = TRUE;
				break;

			case '?':
				fprintf (stderr, "illegal command line arguments\n\n");
				exit (1);
		}

	if (optind >= argc)
	{
		fprintf (stderr, "Usage: %s ANS_NAME\n", argv[0]);
		exit (1);
	}

	temp = getenv ("KQML_HOME");
	if (temp == NULL)
	{
		fprintf (stderr, "KQML_HOME is not defined.\n");
		exit (1);
	}
	if (chdir (temp) <= -1)
	{
		fprintf (stderr, "Could not chdir to %s\n", temp);
		exit (1);
	}
	if (chdir ("adm") <= -1)
	{
		fprintf (stderr, "Could not chdir to %s/adm\n", temp);
		exit (1);
	}

	if ( kqml_initialize (argv[optind], ans_name_lookup) < 0)
	{
		fprintf (stderr, "Unable to initialize router for the ANS.\n");
		exit (1);
	}

#ifdef DEBUG
	fprintf (stderr, "About to do knet_init.\n");
#endif

	/* initialize the cache of agent information */
	if ( init_knet_table () == -1 )
	{
		fprintf (stderr, "Fac : Error in Initialising the table of entries\n");
		exit (1);
	}
  
	/* register an ask-one handler */
#ifdef DEBUG
	fprintf (stderr, "About to register ask_one_handler for 'ask-one'.\n");
#endif

	if ( (kqml_register (kqml_admin_handler, "kqml-admin", 1)) < 0)
	{
		fprintf (stderr, 
			"Unable to register kqml_admin_handler for 'kqml-admin'.\n");
		exit (1);
	}

	if ( (kqml_register (ask_one_handler, "ask-one", 1)) < 0)
	{
		fprintf (stderr, "Unable to register ask_one_handler for 'ask-one'.\n");
		exit (1);
	}

	/* register an ask-all handler */
#ifdef DEBUG
	fprintf (stderr, "About to register ask_all_handler for 'ask-all'.\n");
#endif

	if ( (kqml_register (ask_all_handler, "ask-all", 1)) < 0)
	{
		fprintf (stderr, "Unable to register ask_all_handler for 'ask-all'.\n");
		exit (1);
	}

#ifdef DEBUG
  fprintf (stderr, "About to register register_handler for 'register'.\n");
#endif

	/* register a register handler */
	if ( (kqml_register (register_handler, "register", 1)) < 0)
	{
		fprintf (stderr, "Unable to register register_handler for 'register'.\n");
		exit (1);
	}

#ifdef DEBUG
	fprintf (stderr, "About to register unregister_handler for 'unregister'.\n");
#endif

	/* register an unregister handler */
	if ( (kqml_register (unregister_handler, "unregister", 1)) < 0)
	{
		fprintf (stderr, "Unable to register unregister_handler for 'unregister'.\n");
		exit (1);
	}

#ifdef DEBUG
	fprintf (stderr, "Going into wait loop.\n");
#endif

	/* wait for something to happen */
	while (1)
	{
		sleep (1000);
	}

	fprintf (stderr, "ANS is finished, exiting.\n");
	exit (1);
}

/* find_exact_agent
 * 
 * Find an agent in the table. Used by register and unregister
 *
*/
static struct knet *
find_exact_agent (char *name, char *protocol)
{
	struct knet *node;
	struct net_list *net;

	if (knet_table)
	{
		node = knet_table;
		do
		{
#			ifdef DEBUG
	  			fprintf (stderr, 
					"  Compare '%s' to '%s'\n", node->name, name);
#			endif

			/* Compare only the agent name (do not use the fully qualified
				agent name */
			if (kqml_stricmp (node->name, name) == 0)
			{
				node->net_match = NULL;

				if (protocol && node->net_list)
				{
					net = node->net_list;
					do
					{
						if (kqml_stricmp (net->protocol, protocol) == 0)
						{
							node->net_match = net;
							break;
						}
						else
						{
							net = net->next;
						}
					}
					while (net != node->net_list);
				}

				return node;
			}
			else
			{
				node = node->next;
			}
		}
		while (node != knet_table);
	}

	return NULL;
}

static struct net_list *
find_any_proto (struct knet *node, char *proto, int multi_flag, int *match_len)
{
	struct net_list *temp, *entries;
	char *str;
	int len, any_flag = FALSE;

	if (match_len)
	{
		*match_len = 0;
	}

	if (proto == NULL || proto[0] == '\0' || proto[0] == '*' ||
		kqml_stricmp (proto, "t") == 0 || 
		kqml_stricmp (proto, "nil") == 0 || 
		kqml_stricmp (proto, "null") == 0)
	{
		any_flag = TRUE;
	}

	entries = NULL;

	/* proto may be NULL */
	if (proto && proto[0] == '(')
	{
		/* A list of protocols */
		proto++;
		str = strtok (proto, " \t\n)");
		while (str)
		{
			temp = find_any_proto (node, str, multi_flag, &len);
			if (temp)
			{
				temp->match = entries;
				entries = temp;
				if (match_len)
				{
					*match_len += len;
				}
				if (! multi_flag)
				{
					break;
				}
			}
			str = strtok (NULL, " \t\n)");
		}
	}
	else
	{
		if (node->net_list)
		{
			temp = node->net_list->prev;

			do
			{
#			ifdef DEBUG
	  			fprintf (stderr, 
					"  Compare '%s' to '%s'\n", temp->protocol, proto);
#			endif

				if (any_flag || kqml_stricmp (proto, temp->protocol) == 0)
				{
					temp->match = entries;
					entries = temp;
					if (! multi_flag)
					{
						if (match_len)
						{
							*match_len = strlen (temp->protocol) + 
								strlen (temp->addr_info) + 4;
						}
						break;
					}
					else
					{
						if (match_len)
						{
							*match_len += strlen (temp->protocol) + 
								strlen (temp->addr_info) + 4;
						}
						if (! any_flag)
						{
							/* No point in continuing as there will not
								be any more matches */
							break;
						}
					}
				}
				temp = temp->prev;
			}
			while (temp != node->net_list->prev);
		}
	}

	return entries;
}

/* find_any_agent
 * 
 * Find any matching agent in the table. Used by ask-one and ask-all
 *
 */
static struct knet *
find_any_agent (char *name, char *protocol, int multi_flag, int *match_len)
{
	struct knet *entries, *node;
	struct net_list *net_list;
	int len, any_flag = FALSE;
  
#	ifdef DEBUG
		fprintf (stderr, "Here in find_any_agent with knet_tbl_cnt=%d.\n", 
			knet_tbl_cnt);
#	endif

	if (match_len)
	{
		*match_len = 0;
	}

	if (name == NULL || name[0] == '\0' || name[0] == '*' ||
		kqml_stricmp (name, "t") == 0 ||
		kqml_stricmp (name, "null") == 0 || kqml_stricmp (name, "nil") == 0)
	{
		any_flag = TRUE;
	}

	entries = NULL;
	/* Now find all the entries that match */
	if (knet_table)
	{
		node = knet_table->prev;
		do
		{
#			ifdef DEBUG
	  			fprintf (stderr, 
					"  Compare '%s' to '%s'\n", node->name, name);
#			endif

			/* FOR Heirarchical ANS, handle FQAN and AN */
			if (any_flag || kqml_stricmp (node->name, name) == 0)
			{
				net_list = find_any_proto (node, protocol, multi_flag, &len);
				node->net_match = net_list;
				if (net_list != NULL)
				{
					node->match = entries;
					entries = node;

					if (! multi_flag)
					{
						if (match_len)
						{
							*match_len = strlen (node->name) + len + 2;
						}
						break;
					}
					else
					{
						if (match_len)
						{
							*match_len += strlen (node->name) + len + 3;
						}
					}
				}

				if (! any_flag)
				{
					/* No point in going further, as there will not
						be any more matches */
					break;
				}
			}
			node = node->prev;
		}
		while (node != knet_table->prev);
	}

	return (entries);
}

static char *
form_agents_list (struct knet *entries, int multi_flag, int length)
{
	char *res, *temp;

#	ifdef DEBUG
		fprintf (stderr, "Here in form_agents_list.\n");
#	endif

	if (entries == NULL)
	{
		return NULL;
	}

	/* 
	 * NOTE: When ask-one and ask-all use the same format (see TODO)
	 * ask-one :content (agent_name (protocol addr))
	 * ask-all :content ((agent_name (protocol addr)))
	 * uncomment the code's marked with uncomment
	*/

	if (multi_flag)
	{
		length += 2; /* for the extra paren's */
	}

	/* 
	 *	We will be wasting space, till ask-one and ask-all use the same format 
	 */
	res = malloc (length + 1);

	if (! res)
	{
		fprintf (stderr, "Unable to allocate memory to form entries list.\n");
		return (NULL);
	}

	temp = res;

	/* 
	 *	Comment the following if block when ask-one and ask-all follow the
	 * same format 
	*/
	if (! multi_flag)
	{
		*temp++ = '"';
		temp = memccpy (temp, entries->name, 0, length);
		*(temp - 1) = ':';
		temp = memccpy (temp, entries->net_match->addr_info, 0, length);
		*(temp - 1) = '"';
		*temp = '\0';

		return res;
	}

	if (multi_flag)
	{
		*temp++ = '(';
	}

	while (entries)
	{
		if (multi_flag)
		{
			*temp++ = ' ';
		}
		*temp++ = '(';
		temp = memccpy (temp, entries->name, 0, length);
		temp -= 1;
		while (entries->net_match)
		{
			*temp++ = ' ';
			*temp++ = '(';
			temp = memccpy (temp, entries->net_match->protocol, 0, length);
			*(temp - 1) = ' ';
			temp = memccpy (temp, entries->net_match->addr_info, 0, length);
			*(temp - 1) = ')';
			entries->net_match = entries->net_match->match;
		}
		*temp++ = ')';
		entries = entries->match;
	}

	if (multi_flag)
	{
		*temp++ = ')';
	}

	*temp = '\0';

#	ifdef DEBUG
  		fprintf (stderr, "Leaving form_agents_list with '%s'\n", res);
#	endif

	return (res);
}
