:- library_directory('/usr/local/lib/sicstus2.1/library').
:- use_module(library(lists)).

:- load('$KQML_HOME/lib/interface.ql').
%% :- consult ('$KQML_HOME/src/prolog/interface.pl').

usage:-
	format("~n~n~n **********      Instructions       **************~n~n", []),
	format("Call the predicate goal/2 with agentName and mode (runtime or interpreter)~n", []),
	format("Once the agent is ready, you will get the |: prompt, ~n", []),
	format("where you can key in your queries ~n", []),
	format("You can use the test/0 predicate to key in your msg.~n", []),
	format("~n YOU CAN PRINT THIS MESG. AGAIN BY calling usage/0~n~n~n", []).

%% NOTE:
%% kqmlTermToStr and kqmlStrToTerm are mainly useful only for content
%% and that too only if the language is prolog
%%
test:-
	format("Performative: ", []),
	getAtom(Perf),
	pkril:kqmlAgentName(Sender),
	pkril:kqmlBuildMsg(Perf, TempHandle1, [':sender'], [Sender]),
	%% Read the content. Can be prolog
	format("Content [Press . to terminate input]:", []),
	read(Content),
	pkril:kqmlTermToStr(Content, ContentStr),
	pkril:kqmlPutField(':content', TempHandle1, ContentStr, TempHandle2),
	%% Read the other parameters
	get_keyword_value_pair(ParameterList, ValueList),
	pkril:kqmlPutField(ParameterList, TempHandle2, ValueList, NewHandle),
	%% Send it out
	!, 
	pkril:kqmlSendMsg(NewHandle, ReplyHandle, RetValue),
	%% Print out the response
	printResponse(RetValue, ReplyHandle).

printResponse(_, nil):-
	format(user_error, "Received a NULL response~n", []).

printResponse(0, Handle):-
	format(user_error,"Received the following error~n", []),
	pkril:kqmlPrintMsg(user_error, Handle).

printResponse(1, Handle):-
	format(user_error, "Received the following reply~n", []),
	pkril:kqmlPrintMsg(user_error, Handle).

get_keyword_value_pair([Parameter | FieldList], [ValueStr | ValueList]):-
	format("Parameter: ", []),
	getAtom(Parameter),
	Parameter \== '',
	format("Value:", []),
	getString(ValueStr),
	get_keyword_value_pair(FieldList, ValueList).
get_keyword_value_pair([], []).

%% All our handlers assume that the incoming message is a prolog
%% query and convert the content to prolog and try to execute it.
askif_handler(_, Msg, Reply):-
	pkril:kqmlPrintMsg(user_error, Msg),
	pkril:kqmlGetField(':language', Msg, Lang, _, _),
	(Lang \== "prolog" ->
		lang_error(Reply);
		(pkril:kqmlGetField(':content', Msg, ValueStr, _, _),
		pkril:kqmlStrToTerm(ValueStr, Value),
		pkril:kqmlGetField(':reply-with', Msg, ReplyTag, _, _),
		(call(Value) ->
			ReplyValue = "yes";
			ReplyValue = "no"),
		(ReplyTag \== nil ->
			(pkril:kqmlBuildMsg('reply', ReplyHandle, [':content', ':in-reply-to'],
				[ReplyValue, ReplyTag]),
			Reply = [ReplyHandle]);
			Reply = nil))).
			
askone_handler(_, Msg, Reply):-
	pkril:kqmlPrintMsg(user_error, Msg),
	pkril:kqmlGetField(':language', Msg, Lang, _, _),
	(Lang \== "prolog" ->
		lang_error(Reply);
		(pkril:kqmlGetField(':content', Msg, ValueStr, _, _),
		pkril:kqmlStrToTerm(ValueStr, Value),
		pkril:kqmlGetField(':reply-with', Msg, ReplyTag, _, _),
		(call(Value) ->
			pkril:kqmlTermToStr(Value, ReplyValue); %% The instantiated Value
			ReplyValue = "no"),
		(ReplyTag \== nil ->
			(pkril:kqmlBuildMsg('reply', ReplyHandle, 
				[':content', ':in-reply-to'], [ReplyValue, ReplyTag]),
			Reply = [ReplyHandle]);
			Reply = nil))).

askall_handler(_, Msg, Reply):-
	pkril:kqmlPrintMsg(user_error, Msg),
	pkril:kqmlGetField(':language', Msg, Lang, _, _),
	(Lang \== "prolog" ->
		lang_error(Reply);
		(pkril:kqmlGetField(':content', Msg, ValueStr, _, _),
		pkril:kqmlStrToTerm(ValueStr, Value),
		pkril:kqmlGetField(':reply-with', Msg, ReplyTag, _, _),
		(call(bagof(Value, Value, List)) ->
			pkril:kqmlTermToStr(List, ReplyValue); %% The instantiated Value List
			ReplyValue = "no"),
		(ReplyTag \== nil ->
			(pkril:kqmlBuildMsg('reply', ReplyHandle, 
				[':content', ':in-reply-to'], [ReplyValue, ReplyTag]),
			Reply = [ReplyHandle]);
			Reply = nil))).

tell_handler(_, Msg, Reply):-
	pkril:kqmlPrintMsg(user_error, Msg),
	pkril:kqmlGetField(':language', Msg, Lang, _, _),
	(Lang \== "prolog" ->
		lang_error(Reply);
		(pkril:kqmlGetField(':content', Msg, ValueStr, _, _),
		pkril:kqmlStrToTerm(ValueStr, Value),
		pkril:kqmlGetField(':reply-with', Msg, ReplyTag, _, _),
		(call(myasserta(Value)) ->
			ReplyValue = "yes";
			ReplyValue = "no"),
		(ReplyTag \== nil ->
			(pkril:kqmlBuildMsg('reply', ReplyHandle, 
				[':content', ':in-reply-to'], [ReplyValue, ReplyTag]),
			Reply = [ReplyHandle]);
			Reply = nil))).

myasseta([Head | Tail]):-
	asserta(Head),
	myasserta(Tail).
myasserta([]).
myasserta(Value):-
	asserta(Value).

untell_handler(_, Msg, Reply):-
	pkril:kqmlPrintMsg(user_error, Msg),
	pkril:kqmlGetField(':language', Msg, Lang, _, _),
	(Lang \== "prolog" ->
		lang_error(Reply);
		(pkril:kqmlGetField(':content', Msg, ValueStr, _, _),
		pkril:kqmlStrToTerm(ValueStr, Value),
		pkril:kqmlGetField(':reply-with', Msg, ReplyTag, _, _),
		(call(retractall(Value)) ->
			ReplyValue = "yes";
			ReplyValue = "no"),
		(ReplyTag \== nil ->
			(pkril:kqmlBuildMsg('reply', ReplyHandle, 
				[':content', ':in-reply-to'], [ReplyValue, ReplyTag]),
			Reply = [ReplyHandle]);
			Reply = nil))).

lang_error(Reply):-
	pkril:kqmlBuildMsg('error', ReplyHandle, [':content'], 
		["Language not supported"]),
	Reply = [ReplyHandle].

goal(AgentName, Mode):-
	format("~nStarting Agent ~k~n~n", [AgentName]),
	pkril:kqmlInitialize(AgentName),
	pkril:kqmlRegister(user:askif_handler, 'ask-if', 1),
	pkril:kqmlRegister(user:askall_handler, 'ask-all', 1),
	pkril:kqmlRegister(user:askone_handler, 'ask-one', 1),
	pkril:kqmlRegister(user:tell_handler, 'tell', 1),
	pkril:kqmlRegister(user:untell_handler, 'untell', 1),
	pkril:kqmlInfo(Name, ANSName, Protocol, AddrInfo),
	format("~nAgent ~k using ~k @ ~k and is connected with ~k ANS~n", 
		[Name, ANSName, Protocol, AddrInfo]),
	(Mode == runtime ->
		pkril:kqmlInterpreter;
		true).

getAtom(Atom):-
	getString(String),
	name(Atom, String).

getString(String):-
	getString(String, []).

getString(String, Acc):-
	getChar(X),
	append(Acc, [X], Acc2),
	getString(String, Acc2).

getString(String, String).

getChar(X):-
	get0(X),
	X \== 10,
	X \== -1.

:- usage.
