%e 15000
%p 15000
%a 15000
%o 15000
%k 15000
%n 15000

whitechar       [ \t\n\r]
whitespace      {whitechar}+
digit           [0-9]
alpha           [a-zA-Z]
special         [-<>=+*/&^~_@$%:.!?]
wordchar        ({alpha}|{digit}|{special})
word            {wordchar}+
stringhdr       "#"{digit}+\"
starttoken      "!@#$%"

%{
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <malloc.h>

#include <kqml/kqmlparse.h>

#include <kqml/kqml.tab.h>

#if defined (DEBUG)
#	if !defined (FLEX_SCANNER)
#		define LEXDEBUG 1
#	else
#		define YYDEBUG 1
#	endif
#endif

#if !defined (FLEX_SCANNER)
#	undef input
#	define input() ( \
		((yytchar = yysptr > yysbuf ? U(*--yysptr) : _kqml_parse_get_char ()) \
			== 10 ? (yylineno++, yytchar) : yytchar) == EOF ? 0 : yytchar)
#else
#	define YY_ALWAYS_INTERACTIVE
#	define YY_INPUT(buf, result, max_size) \
		{ \
			int c = _kqml_parse_get_char (); \
			result = (c == EOF ? 0 : 1); \
			buf[0] = (char) c; \
		}
#endif

#define TRUE (1)
#define FALSE (0)

extern char *yylval;

static int first_value (void);
static int check_for_sufficient_space (int extra_space);
static void add_value_token (int current_len);
static void read_bounded_token (int current_len);
static void read_string_token (int current_len);
%}

%%

{whitespace}            { if (! first_value ())
                          {
                             /* don't add leading whitespace to the value
                                buffer */
									  /* Force the whitespace to be space */
									  yytext[0] = ' ';
                             add_value_token (1);
                          }
                        }
{word}                  { yylval = yytext; add_value_token (yyleng); return WORD; }
"("                     { yylval = yytext; add_value_token (yyleng); return '('; }
")"                     { yylval = yytext; add_value_token (yyleng); return ')';  }
","                     { yylval = yytext; add_value_token (yyleng); return ','; }
"`"                     { yylval = yytext; add_value_token (yyleng); return '`'; }
"'"                     { yylval = yytext; add_value_token (yyleng); return '\''; }
"\""							{ yylval = yytext; read_string_token (yyleng); return STRING; }
{stringhdr}             { yylval = yytext; read_bounded_token (yyleng); return BYTE_STRING; }
{starttoken}            { yylval = yytext; return START_TOKEN; }
.                       { yylval = yytext; return INVALID_SYM; }

%%

yywrap ()
{
   return (1);
}

static int recording_value = FALSE;
static int value_len = 0;
static int max_value_len = 0;
static char *value = NULL;

void
_kqml_parse_start_recording_value ()
{
   /* start buffering the value strings assuming that the temp_value_buffer
      will be big enough to contain the entire value */
	value_len = 0;
	recording_value = TRUE;

	if (value == NULL)
	{
		max_value_len = 4096;
		value = malloc (max_value_len);
		
		if (value == NULL)
		{
			max_value_len = 0;
			return;
		}
	}
}

void
_kqml_parse_stop_recording_value (char **s, int *len)
{
	if (value_len > 0)
	{
		*s = value;
		*len = value_len;
	}
	else
	{
		*s = NULL;
		*len = 0;
	}
	recording_value = FALSE;
	value_len = 0;
}

static int
first_value ()
{
	return value_len == 0;
}

static int
check_for_sufficient_space (int extra_space)
{
	char *ptr;
	int new_max_value_len;

	/* If we encountered memory error for this value or if we do not
		have the value buffer return FALSE
	*/
	if (value_len < 0 || value == NULL)
		return FALSE;

   if (value_len + extra_space > max_value_len)
	{
		new_max_value_len = max_value_len;
   	while (value_len + extra_space > new_max_value_len)
   	{
   		new_max_value_len += 512;
   	}

		ptr = realloc (value, new_max_value_len);
		if (ptr == NULL)
		{
			value_len = -1;
			return FALSE;
		}

		value = ptr;
		max_value_len = new_max_value_len;
	}

	return TRUE;
}

static void
add_value_token (int current_len)
{
   if (recording_value)
   {
      if (check_for_sufficient_space (current_len))
		{
      	memcpy (&value[value_len], yytext, current_len);
      	value_len += current_len;
   	}
	}
}

static void
read_bounded_token (int current_len)
{
   int token_len, c;

   if (recording_value)
   {
      sscanf (yytext, "#%d", &token_len);
      if (check_for_sufficient_space (token_len + current_len))
		{
			memcpy (&value[value_len], yytext, current_len);
			value_len += current_len;

			while (token_len)
			{
				c = input ();
				if (c == EOF)
				{
					break;
				}
				value [value_len++] = c;
				token_len--;
			}
   	}
	}
}

static void
read_string_token (int current_len)
{
   int c, c2, increment_len, token_len;

	/* token_len is not used at all */
   if (recording_value)
   {
		token_len = 0;
		increment_len = 256;
      if (check_for_sufficient_space (current_len + increment_len))
		{
			memcpy (&value[value_len], yytext, current_len);
			value_len += current_len;
			token_len += current_len;

			do
			{
				c = input ();
				if (c == EOF)
				{
					break;
				}

				if (increment_len <= 1)
				{
					token_len += 256 - increment_len;
					increment_len = 256;
					if (! check_for_sufficient_space (increment_len))
					{
						break;
					}
				}

				if (c == '\\')
				{
					c2 = input ();
					if (c2 == EOF)
					{
						break;
					}
					value[value_len++] = c;
					value[value_len++] = c2;
					increment_len -= 2;
				}
				else
				{
					value[value_len++] = c;
					increment_len--;
				}
			}
			while (c != '"');
   	}
	token_len += (256 - increment_len);
	}
}
