#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>

#include <kqml.h>

static void Usage (char *progname);

void
main (int argc, char *argv[])
{
	char *string;
   int c, prev_c, i, len, func;
	kqml_message *msg;

	if (argc != 2)
		Usage (argv[0]);

	if (kqml_stricmp (argv[1], "fd") == 0)
		func = 1;
	else if (kqml_stricmp (argv[1], "stream") == 0)
		func = 2;
	else if (kqml_stricmp (argv[1], "buffer") == 0)
	{
		func = 3;
		len = 1024;
		string = malloc (len);
		if (string == NULL)
		{
			printf ("Internal Error. Not enough memory.");
			exit (1);
		}
	}
	else
		Usage (argv[0]);

   while (1)
   {
		if (func == 1)
		{
			i = kqml_parse_fd (0, &msg);
		}
		else if (func == 2)
		{
			i = kqml_parse_stream (stdin, &msg);
		}
		else if (func == 3)
		{
			prev_c = '\0';
			i = 0;
			do
			{
				c = getc (stdin);
				if (i == len)
				{
					len += 1024;
					string = realloc (string, len);
					if (string == NULL)
					{
						printf ("Internal error. Message too big.");
						exit (1);
					}
				}
				
				if (c == EOF)
				{
					string[i++] = '\0';
					break;
				}

				if (c == '\n' && prev_c == '\n')
				{
					string[i++] = '\0';
					break;
				}
				string[i++] = c;
				prev_c = c;
			}
			while (1);

			i = kqml_parse_buffer (string, -1, NULL, &msg);
		}
		else
		{
			printf ("Internal error. Do not know input type.");
			exit (1);
		}

      switch (i) {
		 case KQML_PARSE_TIMEOUT:
			printf ("parser timed out\n");
			break;
       case KQML_PARSE_INCOMPLETE:
         printf ("incomplete message detected\n");
         break;
       case KQML_PARSE_EOF:
         return;
       case KQML_PARSE_OK:
         printf ("\nvalid performative processed\n");
			kqml_print_msg (stdout, msg); 
         break;

       case KQML_PARSE_SYNTAX:
         printf ("syntax error(s) detected in performative\n");
         break;

       case KQML_PARSE_SEMANTIC:
         printf ("semantic error(s) detected in performative\n");
         break;

		 case KQML_PARSE_MEMORY:
			printf ("memory overflow in parsing performative\n");
			break;
      }
   }

	exit (0);
}

static 
void Usage (char *progname)
{
	fprintf (stderr, "Usage: %s {fd | stream | buffer}\n", progname);
	exit (1);
}
